/*
 * \brief  MQTT client
 * \author Timo Nicolai
 * \date   2023-02-23
 */

/*
 * Copyright (C) 2023 gapfruit AG
 */

#pragma once

/* Azure includes */
extern "C" {

#include <azure/az_iot.h>

}

/* Mosquitto includes */
extern "C" {

#include <mosquitto.h>

}

/* local includes */
#include <azure_iot_client/json/json.h>
#include <azure_iot_client/mqtt/check.h>
#include <azure_iot_client/mqtt/mutex.h>
#include <azure_iot_client/mqtt/types.h>

namespace MQTT {
	class Client;
}


class MQTT::Client
{
	public:

		using Callback = void (*)(Topic const &, JSON::Object const &);

	private:

		enum : int
		{
			DEFAULT_PORT      = AZ_IOT_DEFAULT_MQTT_CONNECT_PORT,
			DEFAULT_KEEPALIVE = AZ_IOT_DEFAULT_MQTT_CONNECT_KEEPALIVE_SECONDS,
			SSL_VERIFY_PEER   = 1,
		};

		enum { NUM_SUBSCRIPTIONS_MAX = 10 };

		struct Subscription
		{
			Topic    topic    { };
			Callback callback { };
		};

		Hostname   _server_hostname      { };
		Username   _server_username      { };
		void     (*_server_on_connect)() { nullptr };

		static inline Genode::size_t  _num_clients            { 0 };
		::mosquitto                  *_client                 { nullptr };
		int                           _client_loop_last_error { MOSQ_ERR_SUCCESS };
		bool                          _client_connected       { false };
		mutable Mutex                 _client_connected_mutex { };

		Genode::size_t _num_subscriptions { 0 };
		Subscription   _subscriptions[NUM_SUBSCRIPTIONS_MAX];
		Mutex          _subscriptions_mutex { };

		JSON::Parser _json_parser { };

		static void _on_log(::mosquitto *, void *, int level, char const *message);

		static void _on_connect(::mosquitto *, void *userdata, int rc);

		static void _on_message(::mosquitto *,
		                        void *userdata,
		                        ::mosquitto_message const *message);

		void _configure_mqtt() const;

		void _configure_tls(char const *root_ca_certificates,
		                    char const *device_certificate,
		                    char const *device_private_key) const;

		void _add_subscription(Topic const &topic, Callback callback);

		void _reconnect() const;

		void _reconnect_on_error(int rc, char const *action)
		{
			try {
				check(rc, action);

			} catch (MQTT::Exception const &e) {
				if (e.rc == MOSQ_ERR_NO_CONN || e.rc == MOSQ_ERR_CONN_LOST) {
					{
						Mutex::Guard guard { _client_connected_mutex };

						_client_connected = false;
					}

					_reconnect();
				}

				throw;
			}
		}

	public:

		explicit Client(Client_id const &client_id,
		                char const *root_ca_certificates,
		                char const *device_certificate,
		                char const *device_private_key);
		~Client();

		Client(Client const &)    = delete;
		Client &operator=(Client) = delete;

		void run();

		void connect(Hostname const &hostname,
		             Username const &username,
		             void (*on_connect)());

		void publish(Topic const &topic);
		void publish(Topic const &topic, JSON::Object const &payload_json);

		void subscribe(Topic const &topic, Callback callback);
};
